#include "client.h"

/* stdin and sok use low level read calls
 * stdout uses stdio
 */

int main(int argc, char *argv[])
{
	struct sockaddr_in addr;
	char *buf;
	char ret;

	int dest_port;
	struct hostent *host;
	char *user;
	int keep_alive_secs;

	/* defaults */
	dest_port = 8000;
	host = gethostbyname("192.168.1.1");
	user = getenv("USER");
	keep_alive_secs = (int)(REMOTE_KEEP_ALIVE * 0.80) + 0.5;

	while ((ret=getopt(argc, argv, "p:s:u:k:vh"))!=EOF) {
		switch(ret) {
			case 'p' : dest_port = atoi(optarg);
				   break;
			case 's' : host = gethostbyname(optarg);
				   break;
                        case 'u' : user = (char *)malloc(sizeof(char)*
						   strlen(optarg)+1);
				   /* don't free(user)! see user=getenv */
				   if (NULL==user) {
					   perror("malloc()");
					   return 1;
				   }
				   strcpy(user, optarg);
				   break;
                        case 'k' : keep_alive_secs = atoi(optarg);
                                   break;
			case 'v' : printf("DWUN dial version %s\n", DWUN_VERSION);
				   printf("Copyright (C) 1999 Tim Sutherland\n");
				   printf("See COPYING file in the distribution\n");
				   return 1;
			case 'h' : helpinfo(*argv);
				   return 1;
			default  : printf("Invalid argument\n");
				   helpinfo(*argv);
				   return 1;
		}
	}
	
	/* check for sane values */
	if (0==dest_port) {
		printf("Invalid port: 0\n");
		return 1;
	}
	if (NULL==host) {
		herror("gethostbyname()");
		return 1;
	}
	if (NULL==user) {
		printf("USER not found in environment and not supplied "
				"as option.\n");
		return 1;
	}
	if (keep_alive_secs <= 0) {
		printf("Keep alive timeout must be greater than 0\n");
		return 1;
	}
	/* messages we expect to receive should fit in buf, which is 
	 * MAXLINE+1 length */
	if (strlen(CONMSG) > MAXLINE) {
		printf("CONMSG greater than %d\n", MAXLINE);
		return 1;
	}
	if (strlen(OKAYMSG) > MAXLINE) {
		printf("OKAYMSG greater than %d\n", MAXLINE);
		return 1;
	}
	
	/* add one for terminator */
	buf = (char *)malloc(MAXLINE+1 * sizeof(char));
	if (NULL==buf) {
		perror("malloc()");
		return 1;
	}

	addr.sin_family = AF_INET;
	addr.sin_port = htons(dest_port);
	addr.sin_addr.s_addr = inet_addr(inet_ntoa(*((struct in_addr *)
				host->h_addr)));
	memset(&addr.sin_zero, 0, sizeof(addr.sin_zero));

	if (network(&addr)) { /* socket(), connect() */
		free(buf);
		return 1;
	}

	/* TODO: parse input from server */

	/* REMOTE: hello */
	do {
		if (recvline(buf, strlen(CONMSG))) {
			free(buf);
			return 1;
		}
	} while (strcmp(buf, CONMSG));
	
	/* send user/pass */
	if (auth(user, buf)) {
		free(buf);
		return 1;
	}
	
	/* quit upon 'q' on stdin or timeout */
	if (keep_alive(buf, keep_alive_secs)) {
		free(buf);
		close(sok);
		if (-1==fcntl(0, F_SETFL, 0))
			perror("Unblocking stdin");
		return 1;
	}
	if (-1==fcntl(0, F_SETFL, 0))
		perror("Unblocking stdin");
	free(buf);
	close(sok);
	return 0;
}

/* every keep_alive_secs, send an "alive" message and expect a "ready" back.
 * quit on "q" from stdin
 * print messages from server
 */
int keep_alive(char *buf, int keep_alive_secs)
{
#ifdef HAS_SETITIMER
	struct itimerval itime;
#endif
	struct sigaction cursig;
	fd_set sdfd;
	fd_set sdfdtmp;
	int tmp;

	if (-1==fcntl(0, F_SETFL, O_NONBLOCK)) {
		perror("Setting stdin non-blocking");
		return 1;
	}
	if (-1==fcntl(sok, F_SETFL, O_NONBLOCK)) {
		perror("Setting sok non-blocking");
		return 1;
	}
	
	memset(&cursig, 0, sizeof(cursig));
	cursig.sa_handler = sighandler;
	if (-1==sigaction(SIGALRM, &cursig, NULL)) {
		perror("sigaction() failed");
		return 1;
	}
#ifdef HAS_SETITIMER
	memset(&itime, 0, sizeof(itime));
	itime.it_interval.tv_sec = keep_alive_secs;
	itime.it_value.tv_sec = keep_alive_secs;
	if (-1==setitimer(ITIMER_REAL, &itime, 0)) {
		perror("setitimer() failed");
		return 1;
	}
#else
	alarm(keep_alive_secs);
#endif
	memset(&sdfd, 0, sizeof(&sdfd));
	FD_SET(0, &sdfd);
	FD_SET(sok, &sdfd);
	do {
		memcpy(&sdfdtmp, &sdfd, sizeof(sdfd));
		tmp = select(sok+1, &sdfdtmp, NULL, NULL, NULL);
		if (-1==tmp) {
			if (EINTR!=errno) {
				perror("select()");
				return 1;
			}
		} else if (0==tmp) {
			printf("Portability bug: SIGALRM didn't disrupt system "
					"call or call was restarted\n");
			return 1;
		}
		if (keep_alive_time) {
			keep_alive_time = 0;
#ifndef HAS_SETITIMER
			alarm(keep_alive_secs);
#endif
			if (sendline("alive\n")) {
				return 1;
			}
		}
		if (FD_ISSET(0, &sdfdtmp)) {
			if (flush_recv()) {
				return 1;
			}
		}
		if (FD_ISSET(sok, &sdfdtmp)) {
			if (recvline(buf, MAXLINE)) {
				return 1;
			}
			if (strcmp(buf, "OKAY\n"))
				write(STDOUT_FILENO, buf, strlen(buf));
		}
	} while (1);
}

/* switching to stdio call for buffering */
int sendline(const void *msg)
{
	if (-1==send(sok, msg, strlen(msg), 0)) {
		perror("send()");
		return 1;
	}
	
	return 0;
}

/* returns line of maxlength len, terminating on '\n' or on length */
int recvline(char *buf, int len)
{
	int ret;
	int i;
	/* assumes buf is no longer than MAXLINE */
	if (len > MAXLINE) {
		printf("bug: message to recv() is larger than buf\n");
		return 1;
	}

	i = 0;
	do {
		/* resource temporarily unavaliable: non-blocking */
		ret = recv(sok, buf+i, 1, 0);
		if (0==ret) {
			*(buf+i) = '\0';
			return 1; /* peer left */
		} else if (-1==ret) {
			*(buf+i) = '\0';
			if (EWOULDBLOCK==errno) {
				return 0;
			} else if (EINTR==errno) {
				/* SIGALRM received, select will show there is
				 * still data so keep_alive_time will be checked
				 */
				return 0;
			} else {
				perror("recv()");
				return 1;
			}
		}
		i++;
		if (i>len) {
			*(buf+i+1) = '\0';
			return 0; /* line too long */
		}
	} while ('\n'!=*(buf+i-1));

	*(buf+i) = '\0'; /* +1 to leave the '\n' in */
	return 0; /* newline encountered */
}

int auth(const char *user, char *buf)
{
	char *pass;

	if (sendline("USER ")) {
		return 1;
	}
	if (sendline(user)) {
		return 1;
	}
	if (sendline("\n")) {
		return 1;
	}
	if (recvline(buf, strlen(OKAYMSG))) {
		return 1;
	}
	if (0!=strncmp(buf, OKAYMSG, strlen(OKAYMSG))) {
		printf("Invalid command\n");
		return 1;
	}

	pass = getpass("Password:");
	if (sendline("PASS ")) {
		return 1;
	}
	if (sendline(pass)) {
		memset(pass, 0, strlen(pass));
		return 1;
	}
	memset(pass, 0, strlen(pass));
	if (sendline("\n")) {
		return 1;
	}

	if (sendline("CONNECT\n")) {
		return 1;
	}
	/* server returns OKAY on valid auth */
	if (recvline(buf, strlen(OKAYMSG))) {
		return 1;
	}
	if (0!=strncmp(buf, OKAYMSG, strlen(OKAYMSG))) {
		printf("Invalid username/password\n");
		return 1;
	}
	printf("Authenticated\n");
	return 0;
}

int network(struct sockaddr_in *addr)
{
	sok = socket(AF_INET, SOCK_STREAM, 0);
	if (-1==sok) {
		perror("socket()");
		return 1;
	}
	
	if (-1==connect(sok,(struct sockaddr *)addr, sizeof(struct sockaddr))){
		perror("connect()");
		return 1;
	}
	return 0;
}

void helpinfo(char *progname)
{
	printf("Usage: %s [OPTION] ...\n\n", progname);
	printf("-p\tserver port\n");
	printf("-s\tserver address\n");
	printf("-u\tusername\n");
	printf("-k\tsend keepalive every ... seconds (set to less than servers expectations)\n");
	printf("-v\tversion\n");
	printf("-h\tthis help\n");
	return;
}

void sighandler(int sig)
{
	if (SIGALRM==sig) {
		keep_alive_time = 1;
	}
	return;
}

int flush_recv(void)
{
	int tmp;
	char buf;
	
	tmp = 0; buf = '\0';
	do {
		tmp = read(0, &buf, 1);
		if (-1==tmp) {
			if (EAGAIN==errno) {
				return 0; /* no more data to read */
			} else {
				perror("Network error");
				return 1 ;
			}
		} else if ('q'==buf) {
			(void)sendline("QUIT\n");
			return 1;
		}
	} while (1==tmp);
	perror("Connection reset by peer");
	return 1;
}
