#include "bool.h"
#include "shared.h"
#include "ipc.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <limits.h>
#include <errno.h>
#include "socket.h"
#include "main.h"
#include "util.h"

static bool pipe_send(int pwrite, struct pipe_msg_buf msg);

volatile static sig_atomic_t received_sigusr1 = 0;

bool writeout_authfile(char *shmread)
{
	FILE *fp;

	if (!shmread) {
		return TRUE;
	}
	if (*shmread=='\0') {
		/* bug in make_shmem, couldn't fill shmread with data and 
		 * shmread not set to NULL */
		notice_debug("shmread is empty in writeout_authfile()");
		return TRUE;
	}
	if (!(fp = fopen(set.authfile, "w"))) {
		notice("Couldn't open authfile for writing\n");
		return TRUE;
	}
	fprintf(fp, "%s", shmread);
	if (EOF==(fclose(fp))) {
		notice("Error closing authfile after write\n");
		return TRUE;
	}
	return FALSE;
}
	
/* sets a shared memory segment containing the authfile contents */
int make_shmem(char **shmread, char **shmwrite)
{
	key_t key;
	int shmid;
	struct stat buf;
	int fd;

	*shmread = NULL;
	*shmwrite = NULL;
	
	if (-1==(fd = open(set.authfile, O_RDONLY))) {
		notice_err("Couldn't open authfile %s", set.authfile); 
		return -1;
	}
	if (-1==fstat(fd, &buf)) {
		notice("Couldn't get authfile status %s", set.authfile);
		return -1;
	}
	if ((int)buf.st_size==0) {
		notice("Couldn't read authfile: %s: 0 bytes long\n", 
				set.authfile);
		return -1;
	}
	if (-1==(key = make_key('a'))) {
		return -1;
	}
	if (-1==(shmid = shmget(key, (int)buf.st_size, IPC_CREAT | IPC_EXCL | 
					0600 ))) {
		notice_err("Couldn't get shared memory");
		return -1;
	}
	if (!(*shmwrite = shmat(shmid, 0, 0))) {
		notice_err("Couldn't attach shared memory");
		shmctl(shmid, IPC_RMID, NULL);
		return -1;
	}
	if (!(*shmread = shmat(shmid, 0, SHM_RDONLY))) {
		notice_err("Couldn't attach shared memory");
		shmdt(*shmwrite);
		shmctl(shmid, IPC_RMID, NULL);
		*shmwrite = NULL;
		return -1;
	}
	/* segment will be physically deleted when last reference shmdt()'ed */
	/* shmctl(shmid, IPC_RMID, NULL); */
	/* nope: admin needs to get new reference */

	if ((ssize_t)buf.st_size!=read(fd, *shmwrite, (size_t)buf.st_size)) {
		notice_err("Reading authfile %s", set.authfile);
		shmdt(*shmwrite);
		shmdt(*shmread);
		shmctl(shmid, IPC_RMID, NULL);
		*shmwrite = NULL;
		*shmread = NULL;
		return -1;
	}
	if (close(fd)) {
		notice_err("Closing authfile %s", set.authfile);
		shmdt(*shmwrite);
		shmdt(*shmread);
		shmctl(shmid, IPC_RMID, NULL);
		*shmwrite = NULL;
		*shmread = NULL;
		return -1;
	}
	return shmid;	
}


void remove_shmem(char *shmread, char *shmwrite, int shmid)
{
	shmdt(shmread);
	memset(shmwrite, 0, strlen(shmwrite));
	shmdt(shmwrite);
	shmctl(shmid, IPC_RMID, NULL);
	return;
}

key_t make_key(char c)
{
	key_t key;

	key = ftok(".", c);
	if (-1==key || 0xffffffff==key) {
		notice_err("Creating IPC key");
		return -1;
	}
	return key;
}

void send_clients_list(int pwrite, struct clients *first)
{
	struct pipe_msg_buf msg;
	int shmid;
	struct adminlist *shmpt;
	struct clients *tmp;
	int num_clients;
	int i;
	key_t shmem_key;
	struct sigaction cursig;
	sigset_t waitset;
	sigset_t oldmask;

	/* setup action for SIGUSR1 */
	memset(&cursig, 0, sizeof(cursig));
	cursig.sa_handler = clients_list_handler;
	if (-1==sigaction(SIGUSR1, &cursig, NULL)) {
		notice_err("Signal setup failed");
		return;
	}
	sigemptyset(&waitset);
	sigaddset(&waitset, SIGUSR1);

	/* get some shared memory and fill it with clients stuff */
	for (num_clients=0, tmp=first; tmp; tmp=tmp->next, ++num_clients);
	if (-1==(shmem_key = make_key('c'))) {
		notice("Couldn't create IPC key for admin\n");
		return;
	}
	/* +1 for terminator */
	if (-1==(shmid=shmget(shmem_key, num_clients+1*sizeof(struct adminlist),
					IPC_CREAT | IPC_EXCL | 0600))) {
		notice_err("Couldn't get shared mem for admin");
		return;
	}
	if (!(shmpt = shmat(shmid, 0, 0))) {
		notice_err("Couldn't attach shared mem for admin");
		shmctl(shmid, IPC_RMID, NULL);
		return;
	}
	tmp = first;
	for (i=0; i<num_clients; i++) {
		close(tmp->sok); /* we inherit this */
		if (pwrite!=tmp->pwrite) {
			close(tmp->pwrite);
		}
		(shmpt+i)->pid = tmp->pid;
		(shmpt+i)->connected = tmp->connected;
		strncpy((shmpt+i)->username, tmp->username, USER_LEN);
		strncpy((shmpt+i)->peername, tmp->peername, 256);
		tmp = tmp->next;
	}
	memset(shmpt+i, 0, sizeof(struct adminlist)); /* terminate */
		
	/* send admin child shmid so they can shmget */
	msg.from = getpid();
	msg.command = shmid;
	if (pipe_send(pwrite, msg)) {
		notice("Sending message failed\n");
		shmdt(shmpt);
		shmctl(shmid, IPC_RMID, NULL);
		return;
	}

	/* wait for SIGUSR1 (from adminchild) before deleting shared mem */
	sigprocmask(SIG_BLOCK, &waitset, &oldmask);
	while (!received_sigusr1) {
		sigsuspend(&oldmask);
	}
	sigprocmask(SIG_UNBLOCK, &waitset, NULL);
	
	shmdt(shmpt);
	shmctl(shmid, IPC_RMID, NULL);
	return;
}

void clients_list_handler(int signo)
{
	if (signo==SIGUSR1) {
		received_sigusr1 = 1;
		notice_debug("send_clients_list received SIGUSR1");
	}
}

bool make_pipes(int filedes[2])
{
	if (-1==pipe(filedes)) {
		notice_err("Creating pipe failed");
		return TRUE;
	}
	if ((bool)fcntl(filedes[0], F_SETFL, O_NONBLOCK)) {
		notice_err("Setting pipe non-blocking failed");
	 	return TRUE;
	}
	if ((bool)fcntl(filedes[1], F_SETFL, O_NONBLOCK)) {
		notice_err("Setting pipe non-blocking failed");
	 	return TRUE;
	}
	return FALSE;
}

bool pipe_send(int pwrite, struct pipe_msg_buf msg)
{
	struct timeval tv;
	tv.tv_sec = 20;
	tv.tv_usec = 0;
	switch(sok_send(pwrite, &tv, &msg, sizeof(msg))) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : notice("pipe_send() failed\n");
				  return TRUE;
		case SOK_TIMEOUT: notice("pipe_send() timeout\n");
				  return TRUE;
	}
	return FALSE;
}
