#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <signal.h>
#include <string.h>
#include "util.h"
#include "network.h"
#include "item.h"
#include "local.h"
#include "main.h"

static int sok_from_pid(pid_t pid, struct clients *first);

bool disconnect(pid_t pppd_pid)
{
	if (pppd_pid > 0) {
		if (-1==kill(pppd_pid, SIGTERM)) {
			notice("SIGTERM to pppd failed\n");
			return TRUE;
		}
		return FALSE;
	} else {
		notice_debug("not disconnecting: pppd_pid == 0");
	}
	return TRUE;
}

/* launch pppd */
pid_t dial(void)
{
	pid_t pid;
	
#ifdef NO_PPPD
	return 1; /* > 0 and won't exit */
#endif
	pid = fork();
	switch(pid) {
		case -1 : notice("fork() failed\n");
			  return 0;
		case  0 : cleanup_pppd();
#ifdef DEBUG
			  if ((bool)execl("./debug-pppd", "debug-pppd", NULL)){
				  
				  notice("launching debug-pppd failed\n");
				  kill(getppid(), SIGHUP);
			  }
			  _exit(0);
#else
			  if ((bool)execl("/usr/sbin/pppd", "pppd", 
						  "nodetach","call","provider", 
						  NULL)) {
				  notice("launching pppd failed\n");
				  kill(getppid(), SIGHUP);
			  }
			  _exit(0);
#endif
		default : break;
	}
	return pid;
}

bool con_state(int *constatus, pid_t *pppd_pid, int pon, struct clients *first,
		pid_t lastdisconnected)
{
	int tmpsok;
#ifdef NO_PPPD
	return FALSE;
#endif
	if (*constatus==CONS_DISCONNECTED) {
		if (pon > 0 && *pppd_pid==0) {
			sigprocmask(SIG_BLOCK, &handledset, NULL);
			if (sighandler_can_dial) {
				sighandler_can_dial = 0;
				(void)alarm(5);
			} else {
				sigprocmask(SIG_UNBLOCK, &handledset, NULL);
				return FALSE;
			}
			sigprocmask(SIG_UNBLOCK, &handledset, NULL);
			*pppd_pid = dial();
			if (*pppd_pid > 0) {
				sendtoallcon("#Connecting\n", first);
				*constatus = CONS_CONNECTING;
				return FALSE;
			} else {
				return TRUE;
			}
		}
		return FALSE;
	}

	if (pon==0){
		if (-1==(tmpsok = sok_from_pid(lastdisconnected, first))) {
			(void)disconnect(*pppd_pid);
		} else {
			if (disconnect(*pppd_pid)) {
				sendalert("#Disconnect failed\n", tmpsok);
			} else {
				sendalert("#Disconnected\n", tmpsok);
			}
		}
		*constatus = CONS_DISCONNECTED; /* if failure give up */
	}
	return FALSE;
}

/* uggh */
void signal_acton(int conaction, pid_t *send_list_pid, pid_t *adminchildpid,
		pid_t *pppd_pid, int *pppd_fail, int *sokcount, int *pon,
		int *constatus, struct clients *start, struct clients **end,
		pid_t lastdisconnected)
{
	int status;
	pid_t pid;
	
	switch(conaction) {
		case ACT_CHILDDEAD:
			if ((pid = waitpid(0, &status, 0)) <= 0) {
				notice_debug("bug in signal_acton");
				break;
			}
			notice_debug("ACT_CHILDDEAD for %d", pid);
			if (pid==*send_list_pid) {
				*send_list_pid = 0;
				break;
			} else if (pid==*pppd_pid) {
				/* *constatus probably already done */
				*constatus = CONS_DISCONNECTED;
				*pppd_pid = 0;
				if (pon > 0) {
					sendtoallcon("#Reconnecting\n",
							start->next);
					if (WIFEXITED(status)) {
						*pppd_fail+=1;
					} else {
						*pppd_fail = 0;
					}
				}
				break;
			} else {
				/* child() */
				if (pid==*adminchildpid) {
					*adminchildpid = 0;
				}
				switch(item_rem(start, end, pid)) {
					case ITEM_NORMAL:
						*sokcount-=1;
						break;
					case ITEM_DISCONNECT:
						*sokcount-=1;
						*pon-=1;
						break;
					case ITEM_NOTFOUND:
						notice_debug("exit of unknown process\n");
						break;
				}
			}
			break;
		case ACT_FATAL:
			*constatus = CONS_DISCONNECTED;
			notice("fatal pppd error\n");
			sendtoallcon("#Connect failed\n", start->next);
			(void)disconnect(*pppd_pid); /* just in case */
			setall_disconnected(start);
			*pon = 0;
			break;
		case ACT_RECONNECT:
			if (*constatus==CONS_DISCONNECTED || pon <= 0) {
				notice_debug("not reconnecting: not connected");
				break;
			}
			notice("reconnecting\n");
			/* sendtoallcon("#Reconnecting\n", start->next); */
			if (disconnect(*pppd_pid)) {
				sendtoall("#Disconnect failed\n", start->next);
			} else {
				/* trust in con_state */
				*constatus = CONS_DISCONNECTED;
			}
			break;
		case ACT_CONNECTED:
			if (*constatus==CONS_CONNECTED) {
				notice_debug("already connected, ignoring signal\n");
				break;
			}
			if (*pppd_pid > 0) {
				notice("connected");
				sendtoallcon("#Connected\n", start->next);
				*constatus = CONS_CONNECTED;
			} else {
				notice_debug("can't be connected - pppd_pid==0");
			}
			break;
		case ACT_CANDIAL:
			if (con_state(constatus, pppd_pid, *pon, start->next,
						lastdisconnected)){
				pppd_fail++;
			}
	}
#ifndef NO_PPPD
	if (*pppd_fail > MAXPPPD_FAIL) {
		notice("pppd failed %d times in a row\n", *pppd_fail);
		*pppd_fail = 0;
		raise(SIGINT); /* ACT_FATAL */
	}
#endif
	return;
}

int sok_from_pid(pid_t pid, struct clients *first)
{
	struct clients *pt;
	if (!first)
		return -1;
	for (pt = first; pt->pid!=pid; pt=pt->next) {
		if (!pt->next)
			return -1;
	}
	return pt->sok;
}
		
