#include <stdio.h>
#include <limits.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/select.h>
#include "shared.h"
#include "rets.h"
#include "network.h"
#include "child.h"
#include "socket.h"
#include "item.h"
#include "local.h"
#include "exec.h"
#include "util.h"
#include "main.h"

static bool client_add(int sd, char *shmread, struct clients **end, 
		int child_pwrite, fd_set *readers);
static void client_drop(int sd, const char *msg);
static int parent(int sd);
static int recv_child(int pread, struct clients *start, int shmid, 
		pid_t *send_list_pid, pid_t *adminchildpid, 
		pid_t *lastdisconnected, int constatus);
static void sighandler(int sig);
static void sighandler_write(int conaction);
static bool sigsetup(void);
static bool pipe_send(int pwrite, struct pipe_msg_buf msg);
static bool pipe_recv(int pread, struct pipe_msg_buf *msg);


volatile sig_atomic_t sighandler_write_pipe;

/* wait for socket connection to us on port */
int server(void)
{
#define BACKLOG 10
	int sd;
	struct sockaddr_in address;
	int optval;

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if(-1==sd) {
		notice_err("socket() failed");
		return QSHUTDOWN;
        }
	address.sin_family = AF_INET;
	address.sin_port = htons(set.port);
	memset(&address.sin_zero, 0, sizeof(address.sin_zero));
	
	/* see RFC 793 for TIME_WAIT. We want to be able to bind to the 
	 * server socket immediately after the server is terminated, so
	 * we enable SO_REUSEADDR */
	optval = 1; /* Doesn't seem necessary in Linux */
	if (-1==setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, 
				(const void *)&optval, sizeof(optval))) {
		notice_err("setsockopt() failed");
		return QSHUTDOWN;
	}
	address.sin_addr.s_addr = set.bind_ip;
	if (-1==bind(sd, (struct sockaddr *)&address, sizeof(address))) {
		notice_err("bind() failed");
		return QSHUTDOWN;
	}
	
	if(-1==listen(sd, BACKLOG)) {
		notice_err("listen() failed");
		return QSHUTDOWN;
	}
	/* accept() may fail, we don't want it to block */
	if (-1==fcntl(sd, F_SETFL, O_NONBLOCK)) {
		notice_err("Setting socket non-blocking failed");
    		return QSHUTDOWN;
	}
	return(parent(sd));
}

int parent(int sd)
{
	/* state */
	int sokcount = 0; 	/* number of clients */
	int pon = 0;		/* number of connected clients */
	int constatus = CONS_DISCONNECTED;
	int pppd_fail = 0;	/* times in a row pppd fails */
	/* tmp vars */
	int ret;
	pid_t pppd_pid = 0;
	pid_t lastdisconnected;	/* notify this person when we disconnect */
	/* for select() */
	fd_set readersbak;
	fd_set readers;
	/* clients linked list */
	struct clients *start;
	struct clients *end;
	/* IPC */
	int pipe_parent[2];
	int shmid;		/* shared memory */
	char *shmwrite;		/* contain authfile contents */
	char *shmread;
	pid_t send_list_pid = 0;/* process that sends clients list to admin */
	pid_t adminchildpid = 0;/* child authed as admin */
	/* signals */
	int pipe_signals[2];
	int conaction;

	time_to_hangup = QNORMAL;
	conaction = ACT_NOTHING;
	sighandler_can_dial = 1;

	/* initialize list */
	if (!(start = (struct clients *)malloc(sizeof(struct clients)))) {
		notice_err("malloc() failed");
		return QSHUTDOWN;
	}
	start->next = NULL;
	end = start;
	
	/* signal handling */
	if (make_pipes(pipe_signals)) {
		return QSHUTDOWN;
	}
	sighandler_write_pipe = dup(pipe_signals[1]);
	if (sigsetup()) {
		notice_err("Setting up signal handlers failed");
		return QSHUTDOWN;
	}
	
	/* child->parent communication */
	if (make_pipes(pipe_parent)) {
		return QSHUTDOWN;
	}
	
	/* make shmread/shmwrite contain authfile */
	if (-1==(shmid = make_shmem(&shmread, &shmwrite))) {
		close(pipe_parent[1]); close(pipe_parent[2]);
		return QSHUTDOWN;
	}
	
#ifndef DEBUG
	if (become_daemon()) {
		time_to_hangup = QSHUTDOWN;
	}
#endif
	
	FD_ZERO(&readersbak);
	FD_SET(sd, &readersbak);
	FD_SET(pipe_parent[0], &readersbak);
	FD_SET(pipe_signals[0], &readersbak);
	
	while (time_to_hangup==QNORMAL) {
		memcpy(&readers, &readersbak, sizeof(readers)); 
		ret = select(pipe_parent[0]+1, &readers, NULL, NULL, NULL);
		if (-1==ret) {
			if (errno!=EINTR) {
				notice_err("select() failed");
				time_to_hangup = QSHUTDOWN;
				continue;
			} else {
				memset(&readers, 0, sizeof(readers));
			}
		}

		/* signals */
		sigprocmask_noint(SIG_BLOCK, &handledset);
		if (FD_ISSET(pipe_signals[0], &readers)) {
			if (read(pipe_signals[0], &conaction, 1) < 1) {
				notice_debug("bad read(pipe_signals)");
				conaction = ACT_NOTHING;
			}
			/* urgh, too much state */
			signal_acton(conaction, &send_list_pid, &adminchildpid,
					&pppd_pid, &pppd_fail, &sokcount, &pon,
					&constatus, start, &end, 
					lastdisconnected);
			conaction = ACT_NOTHING;
		}
		sigprocmask_noint(SIG_UNBLOCK, &handledset);

		/* client trying to connect? */
		if (FD_ISSET(sd, &readers)) {
			if (sokcount >= set.maxcons) {
				client_drop(sd, "max connections exceeded");
			} else {
				if (!client_add(sd, shmread, &end, 
							pipe_parent[1],
							&readersbak)) {
					sokcount++;
				}
			}
		}

		/* message from child */
		if (FD_ISSET(pipe_parent[0], &readers)) {
			pon += recv_child(pipe_parent[0], start, shmid, 
					&send_list_pid, &adminchildpid,
					&lastdisconnected, constatus);
		}
		/* adjust connection state */
		if (con_state(&constatus, &pppd_pid, pon, start->next,
					lastdisconnected)) {
			pppd_fail++;
		}
	} /* while (time_to_hangup==QNORMAL) */
	
	sendtoall("#Server shutting down\n", start->next);
	if (constatus!=CONS_DISCONNECTED) {
		if (disconnect(pppd_pid)) {
			sendtoall("#Disconnect failed\n", start->next);
		}
	}
	killall(start, &end, adminchildpid);
	if (send_list_pid > 0) {
		kill(send_list_pid, SIGUSR1);
	}
	free(start);
	close(pipe_parent[0]); close(pipe_parent[1]);
	writeout_authfile(shmread);
	remove_shmem(shmread, shmwrite, shmid);
	close(sd);
	return time_to_hangup;
}

void sighandler(int sig)
{
	switch(sig) {
		case SIGHUP : time_to_hangup = QSHUTDOWN;
			      sighandler_write(ACT_HANGUP);
			      notice("SIGHUP caught\n");
			      break;
		case SIGINT : sighandler_write(ACT_FATAL);
			      notice_debug("SIGINT caught");
			      break;
		case SIGUSR1: sighandler_write(ACT_CONNECTED);
			      notice_debug("SIGUSR1 caught");
			      break;
		case SIGUSR2: sighandler_write(ACT_RECONNECT);
			      notice_debug("SIGUSR2 caught");
			      break;
		case SIGCHLD: sighandler_write(ACT_CHILDDEAD);
			      notice_debug("SIGCHLD caught");
			      break;
		case SIGPIPE: notice_debug("SIGPIPE caught");
			      break;
		case SIGALRM: sighandler_write(ACT_CANDIAL);
			      sighandler_can_dial = 1;
			      notice_debug("SIGALRM caught");
			      break;
		default     : notice("invalid signal caught\n");
			      break;
	}
	return;
}

void sighandler_write(int conaction)
{
	if (write(sighandler_write_pipe, &conaction, 1) < 1) {
		notice_debug("sighandler_write failed");
	}
	return;
}

/* associate sighandler function with signals */
bool sigsetup(void)
{
	struct sigaction ignore;
	struct sigaction cursig;
	
	memset(&ignore, 0, sizeof(ignore));
	memset(&cursig, 0, sizeof(cursig));
	
	ignore.sa_handler = SIG_IGN;
	cursig.sa_handler = sighandler;
	cursig.sa_flags   = SA_RESTART;
	
	/* block in sighandler */
	sigaddset(&cursig.sa_mask, SIGHUP);
	sigaddset(&cursig.sa_mask, SIGUSR1);
	sigaddset(&cursig.sa_mask, SIGUSR2);
	sigaddset(&cursig.sa_mask, SIGINT);
	sigaddset(&cursig.sa_mask, SIGPIPE);
	sigaddset(&cursig.sa_mask, SIGTERM);
	sigaddset(&cursig.sa_mask, SIGALRM);
	sigaddset(&cursig.sa_mask, SIGCHLD);

	/* FIXME: do we need this? */
	sigemptyset(&handledset);
	/* sigaddset(sig_ignore, SIGHUP); [ allow SIGHUP if we get stuck ] */
	sigaddset(&handledset, SIGUSR1);
	sigaddset(&handledset, SIGUSR2);
#ifndef DEBUG
	sigaddset(&handledset, SIGINT); /* debugger uses SIGINT */
#endif
	sigaddset(&handledset, SIGCHLD);
	sigaddset(&handledset, SIGTERM);
	sigaddset(&handledset, SIGPIPE);
	sigaddset(&handledset, SIGALRM);
	
	if (		-1==sigaction(SIGHUP , &cursig, NULL) ||
			-1==sigaction(SIGUSR1, &cursig, NULL) ||
			-1==sigaction(SIGUSR2, &cursig, NULL) ||
			-1==sigaction(SIGINT , &cursig, NULL) ||
			-1==sigaction(SIGCHLD, &cursig, NULL) ||
			-1==sigaction(SIGPIPE, &cursig, NULL) ||
			-1==sigaction(SIGALRM, &cursig, NULL) ||
			/* pppd can do kill(0, signal we send it [SIGTERM] ) */
			-1==sigaction(SIGTERM, &ignore, NULL)) {
		return TRUE;
	}
	return FALSE;
}

bool sendtoallcon(char *buf, struct clients *first)
{
	while (first!=NULL) {
		if (first->connected) {
			(void)sendalert(buf, first->sok);
		}
		first = first->next;
	}
	return FALSE;
}

bool sendtoall(char *buf, struct clients *first)
{
	while (first!=NULL) {
		(void)sendalert(buf, first->sok);
		first = first->next;
	}
	return FALSE;
}

bool sendalert(char *msg, int sok)
{
	struct timeval tv;
	int ret;
	char *name;
	
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	
	ret = sok_send(sok, &tv, msg, strlen(msg));
	if (ret==SOK_SUCCESS) {
		return FALSE;
	}
	if (!(name = makepeername(sok))) {
		return TRUE;
	}
	switch(ret) {
		case SOK_FAILED : notice_err("sendalert() failed for %s", name);
				  break;
		case SOK_TIMEOUT: notice_err("sendalert() timeout for %s", name);
				  break;
	}
	free(name);
	return TRUE;
}

/* handles messages from children. 
 * returns number pon should be increased by */
int recv_child(int pread, struct clients *start, int shmid, 
		pid_t *send_list_pid, pid_t *adminchildpid,
		pid_t *lastdisconnected, int constatus)
{
	struct pipe_msg_buf msg;
	int pwrite;
	
	if (pipe_recv(pread, &msg)) {
		return 0;
	}
	notice_debug("recv_child message from %d", msg.from);
	if (-1==(pwrite = pipe_from_pid(msg.from, start))) {
		return 0;
	}
	switch(msg.command) {
		case cCON:
			if (!item_con(msg.from, start, TRUE)) {
				msg.from = 1;
				msg.command = constatus;
				if (pipe_send(pwrite, msg)) {
					notice("Sending message failed\n");
				}
				return 1; /* pon+=1 */
			}
			break;
		case cDISCONNECT:
			if (!item_con(msg.from, start, FALSE)) {
				*lastdisconnected = msg.from;
				return -1; /* pon+=-1 */
			}
			break;
		case cAUTH:
			item_auth(msg.from, msg.username, start, TRUE);
			break;
		case cDEAUTH:
			item_auth(msg.from, NULL, start, FALSE);
			break;
		case cADMIN_LIST:
			if (*adminchildpid!=0 && *adminchildpid!=msg.from) {
				msg.from = 1;
				msg.command = -1; /* only one admin allowed */
				if (pipe_send(pwrite, msg)) {
					notice("Sending message failed\n");
				}
				break;
			}
			*adminchildpid = msg.from;
			*send_list_pid = fork();
			switch((int)*send_list_pid) {
				case -1: notice("fork() failed\n");
					 break;
				case  0: cleanup();
					 send_clients_list(pwrite, start->next);
					 _exit(0);
				default: break;
			}
			break;
		case cADMIN_AUTHWRITE:
			/* send shmwrite to child */
			msg.from = 1;
			msg.command = shmid;
			if (pipe_send(pwrite, msg)) {
				notice("Sending message failed\n");
			}
			break;
		case cADMIN_RESTART:
			time_to_hangup = QRESTART;
			break;
		default: 
			notice("Invalid message received\n");
			notice_debug("message = %d", msg.command);
			break;
	}
	return 0;
}

void client_drop(int sd, const char *msg)
{
	struct sockaddr_in address;
	size_t addrlen = sizeof(address);
	int sok;
	char *name;
	
	do {
		sok = accept(sd, (struct sockaddr *)&address, &addrlen);
	} while (sok==-1 && errno==EINTR);
	name = makepeername(sok);
	(void)close(sok);
	
	if (name) {
		notice("%s dropped: %s\n", name, msg);
		free(name);
	}
	return;
}
	
bool client_add(int sd, char *shmread, struct clients **end, 
		int child_pwrite, fd_set *readersbak)
{
	struct sockaddr_in address;
	size_t addrlen = sizeof(address);
	int sok;
	pid_t pid;
	int pipe_child[2];
	char *name;
	
	do {
		sok = accept(sd, (struct sockaddr *)&address, &addrlen);
	} while (sok==-1 && errno==EINTR);
	if (sok==-1) {
		return TRUE; /* reset between select and accept */
	}
	name = makepeername(sok);
	if (ip_out_of_range(name)) {
		notice("connection denied to %s\n", name);
		free(name);
		close(sok);
		return TRUE;
	}
	if (make_pipes(pipe_child)) {
		notice_err("pipe() failed");
		(void)close(sok);
		return TRUE;
	}
	
	pid = fork();
	switch(pid) {
		case -1: notice_err("fork() failed");
			 close(sok);
			 return TRUE;
		case  0: cleanup();
			 close(pipe_child[1]);
			 close((*end)->sok);
			 close((*end)->pwrite);
			 child(sok, child_pwrite, pipe_child[0], shmread);
			 _exit(0);
		default: (void)close(pipe_child[0]);
			 if (item_add(end, pid, sok, pipe_child[1], name)) {
				 (void)close(sok);
				 (void)close(pipe_child[1]);
				 return TRUE;
			 }
	}
	FD_SET(pipe_child[1], readersbak);
	return FALSE;
}

bool pipe_send(int pwrite, struct pipe_msg_buf msg)
{
	struct timeval tv;
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	
	switch(sok_send(pwrite, &tv, &msg, sizeof(struct pipe_msg_buf))) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : notice_err("pipe_send() failed");
				  return TRUE;
		case SOK_TIMEOUT: notice_err("pipe_send() timeout");
				  return TRUE;
	}
	return FALSE;
}

bool pipe_recv(int pread, struct pipe_msg_buf *msg)
{
	struct timeval tv;
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	switch(sok_recv(pread, &tv, msg, sizeof(struct pipe_msg_buf))) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : notice_err("pipe_recv() failed");
				  return TRUE;
		case SOK_TIMEOUT: notice("pipe_recv() timeout\n");
				  return TRUE;
	}
	return FALSE;
}

