#include "bool.h"
#include <syslog.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include "util.h"
#include "child.h"
#include "main.h"

static bool sigaction_noint(int signum, const struct sigaction *act);

/* asprintf/vsnprintf aren't portable. bugger */
void notice(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	if (LOG_TO_SYSLOG) {
		vsnprintf(error_messages, MAX_LOG_LENGTH, format, ap);
		syslog(LOG_WARNING, "%s", error_messages);
	} else {
		vfprintf(stderr, format, ap);
		fflush(stderr);
	}
	va_end(ap);
	return;
}

void notice_err(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	if (LOG_TO_SYSLOG) {
		vsnprintf(error_messages, MAX_LOG_LENGTH, format, ap);
		syslog(LOG_WARNING, "%s: %s", error_messages, strerror(errno));
	} else {
		vfprintf(stderr, format, ap);
		fprintf(stderr, ": %s\n", strerror(errno));
		fflush(stderr);
	}
	va_end(ap);
	return;
}

void notice_debug(const char *format, ...)
{
	va_list ap;
	
	if (!set.debug) {
		return;
	}

	va_start(ap, format);
	if (LOG_TO_SYSLOG) {
		vsnprintf(error_messages, MAX_LOG_LENGTH, format, ap);
		syslog(LOG_DEBUG, "%s: %m", error_messages);
	} else {
		vfprintf(stderr, format, ap);
		fprintf(stderr, ": %s\n", strerror(errno));
		fflush(stderr);
	}
	va_end(ap);
	return;
}

bool become_daemon(void)
{
	FILE *pidfp;

	switch(fork()) {
		case -1: notice_err("fork() failed");
			 return TRUE;
		case  0: break;
		default: exit(0);
	}
	if (-1==setsid()) {
		notice_err("setsid() failed");
		return TRUE;
	}
	switch(fork()) {
		case -1: notice_err("fork() failed");
			 return TRUE;
		case  0: break;
		default: exit(0);
	}

	umask(022);
	if (!(pidfp = fopen(set.dwunpid, "w"))) {
		notice_err("Can't open pid file %s for writing", set.dwunpid);
		return TRUE;
	}

	/* must be after daemon or pid will have changed */
	fprintf(pidfp, "%d", getpid());
	if (-1==fclose(pidfp)) {
		notice_err("fclose() failed");
		return TRUE;
	}

	umask(0);
	close(0);
	close(1);
	close(2);
	freopen("/dev/null", "r", stdin);
	freopen("/dev/null", "w", stdout);
	freopen("/dev/null", "w", stderr);

	openlog("dwun", LOG_PID, LOG_DAEMON);
	syslog(LOG_INFO, "starting");
	LOG_TO_SYSLOG = TRUE;
	return FALSE;
}

/* remove inherited signals from parent, except SIGTERM which will remain
 * ignored */
void cleanup(void)
{
	struct sigaction cursig;
	sigset_t emptymask;

	memset(&cursig, 0, sizeof(cursig));
	cursig.sa_handler = SIG_DFL;
	if (		sigaction_noint(SIGHUP , &cursig) ||
			sigaction_noint(SIGUSR1, &cursig) ||
			sigaction_noint(SIGUSR2, &cursig) ||
			sigaction_noint(SIGINT , &cursig) ||
			sigaction_noint(SIGCHLD, &cursig) ||
			sigaction_noint(SIGPIPE, &cursig)) {
		notice_debug("cleanup of signals failed");
	}
	sigemptyset(&emptymask);
	sigprocmask_noint(SIG_SETMASK, &emptymask);
	return;
}

void cleanup_pppd(void)
{
	struct sigaction cursig;
	sigset_t emptymask;
	
	memset(&cursig, 0, sizeof(cursig));
	cursig.sa_handler = SIG_DFL;
	if (		sigaction_noint(SIGHUP, &cursig)  ||
			sigaction_noint(SIGUSR1, &cursig) ||
			sigaction_noint(SIGUSR2, &cursig) ||
			sigaction_noint(SIGINT, &cursig)  ||
			sigaction_noint(SIGCHLD, &cursig) ||
			sigaction_noint(SIGTERM, &cursig) ||
			sigaction_noint(SIGPIPE, &cursig)) {
		notice_debug("pppd cleanup of signals failed");
	}
	sigemptyset(&emptymask);
	if (sigprocmask_noint(SIG_SETMASK, &emptymask))
		notice_debug("sigprocmask_noint failed in cleanup_pppd");
	return;
}
	
bool sigaction_noint(int signum, const struct sigaction *act)
{
	int ret;
	while (-1==(ret=sigaction(signum, act, NULL)) && errno==EINTR);
	if (ret)
		return TRUE;
	return FALSE;
}

bool sigprocmask_noint(int how, const sigset_t *set)
{
	int ret;
	while (-1==(ret=sigprocmask(how, set, NULL)) && errno==EINTR);
	if (ret)
		return TRUE;
	return FALSE;
}

bool alloc_copy(char **s, char *msg)
{
	if (!(*s = malloc(strlen(msg)+1 * sizeof(char)))) {
		notice_err("malloc() failed");
		return TRUE;
	}
	strcpy(*s, msg);
	return FALSE;
}

/* number of characters in string s before the next character is c (or 0 if
 * not found) */
int substringlen(const char *s, char c)
{
	const char *pt;
	if (!s)
		return 0;
	for (pt = s; *pt; ++pt) {
		if (*pt==c)
			return pt-s;
	}
	return 0;
}
